import argparse


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


class TypeToNone:
    def __init__(self, _type):
        self.type = _type

    def __call__(self, v):
        if v in (None, 'None', 'null'):
            return None
        return self.type(v)

    def __repr__(self):
        return f'TypeToNone({repr(self.type)})'

    @property
    def valid_types(self):
        return self.type, type(None)


str2none = TypeToNone(str)



from .logging import make_sequential_log_dir, get_logger, TapeRecorder
from .core import nested_dict_update
from .namespacify import Namespacify
from .fig import Config

logger = get_logger()
tape = TapeRecorder()



import yaml

from functools import reduce

from expfig.core._str_types import str2none


CMD_LINE_YAML_REPLACEMENTS = [('{', ' {'), (':', ': ')]


class YamlType:
    __name__ = 'YamlType'

    def __init__(self, yaml_default):
        self.yaml_default = yaml_default

    def __call__(self, value):
        if self.yaml_default and is_yaml_obj(value):
            return value

        if is_yaml_str(value):
            try:
                return _load_yaml_value(value)
            except yaml.YAMLError:
                if self.yaml_default:
                    raise

        elif self.yaml_default:
            raise yaml.constructor.ConstructorError(f"value '{value}' does not begin with a yaml tag (!).")

        return str2none(value)


def is_yaml_obj(value):
    return getattr(value, 'yaml_tag', None) is not None


def is_yaml_str(value):
    return isinstance(value, str) and value.startswith('!')


def _load_yaml_value(value):
    try:
        value = yaml.safe_load(value)
    except yaml.YAMLError:
        try:
            value = yaml.safe_load(f'{value} {{}}')
        except yaml.YAMLError:
            value = reduce(lambda _str, kv: _str.replace(*kv), CMD_LINE_YAML_REPLACEMENTS, value)
            value = yaml.safe_load(value)

    return value



from collections import UserDict

from expfig.utils import api


def unflatten(arguments, delimiter='.'):
    if set(arguments.keys()) == {''}:
        return arguments['']

    restructured = {}
    for key, value in arguments.items():
        top_key, _, bottom_keys = key.partition(delimiter)
        update_with = {top_key: unflatten({bottom_keys: value})}
        nested_dict_update(restructured, update_with)

    return restructured


def flatten(nested, delimiter='.', levels=None, *, _key_stack=()):
    if levels is None:
        levels = depth(nested) + 1

    flat = {}
    for k, v in nested.items():
        if api.is_dict_like(v) and (
            (levels >= 0 and len(_key_stack) < levels - 1) or
            (levels < 0 and depth(v) > -1 * levels)
        ):
            flat.update(flatten(v, delimiter, levels=levels, _key_stack=[*_key_stack, k]))
        else:
            flat_key = delimiter.join([*_key_stack, k])
            flat[flat_key] = v

    return flat


def depth(d):
    if api.is_dict_like(d):
        return 1 + (max(map(depth, d.values())) if d else 0)
    return 0


def nested_dict_update(nested_dict, *args, nest_namespacify=False, **kwargs):
    from expfig import Namespacify  # prevent circular import

    if args:
        if len(args) != 1 or not isinstance(args[0], (dict, UserDict)):
            raise TypeError('Invalid arguments')
        elif kwargs:
            raise TypeError('Cannot pass both args and kwargs.')

        d = args[0]
    else:
        d = kwargs

    for k, v in d.items():
        if isinstance(v, (dict, UserDict)):
            if k in nested_dict and api.is_dict_like(nested_dict[k]):
                nested_dict_update(
                    nested_dict[k], v, nest_namespacify=(nest_namespacify or isinstance(nested_dict[k], Namespacify))
                )
            else:
                nested_dict[k] = Namespacify(v) if nest_namespacify else v
        else:
            nested_dict[k] = v

    return nested_dict



from collections.abc import Iterable


def is_dict_like(obj) -> bool:
    """
    Stolen from pandas

    Check if the object is dict-like.

    Parameters
    ----------
    obj : The object to check

    Returns
    -------
    is_dict_like : bool
        Whether `obj` has dict-like properties.

    Examples
    --------
    >>> is_dict_like({1: 2})
    True
    >>> is_dict_like([1, 2, 3])
    False
    >>> is_dict_like(dict)
    False
    >>> is_dict_like(dict())
    True
    """
    dict_like_attrs = ("__getitem__", "keys", "__contains__")
    return (
        all(hasattr(obj, attr) for attr in dict_like_attrs)
        # [GH 25196] exclude classes
        and not isinstance(obj, type)
    )


def is_list_like(obj):
    """
    Stolen from pandas

    Check if the object is list-like.

    Objects that are considered list-like are for example Python
    lists, tuples, sets, NumPy arrays, and Pandas Series.

    Strings and datetime objects, however, are not considered list-like.

    Parameters
    ----------
    obj : The object to check.

    Returns
    -------
    is_list_like : bool
        Whether `obj` has list-like properties.

    Examples
    --------
    >>> is_list_like([1, 2, 3])
    True
    >>> is_list_like({1, 2, 3})
    True
    >>> is_list_like(datetime(2017, 1, 1))
    False
    >>> is_list_like("foo")
    False
    >>> is_list_like(1)
    False
    """

    return (isinstance(obj, Iterable) and
            not isinstance(obj, (str, bytes)))



from difflib import get_close_matches

nlt = '\n\t'

best_matches = dict()


def get_similar_args(unrecognized_args, known_args):
    unrecognized_options = [opt.partition('=')[0].lstrip('-') for opt in unrecognized_args if opt.startswith('--')]

    for bad_opt in unrecognized_options:
        best_matches[f'--{bad_opt}'] = [f'--{s}' for s in get_close_matches(bad_opt, known_args)]

    return best_matches


def get_similar_args_str_fmt(unrecognized_args, known_args):
    similar_args = get_similar_args(unrecognized_args, known_args)

    string = [f"Unrecognized arguments '{unrecognized_args}'"]

    for unrecognized, matches in similar_args.items():
        if matches:
            s = f"'{unrecognized}' is not a recognized argument.\n\nThe most similar arguments are{nlt}{nlt.join(matches)}"
            string.append(s)

    if len(string) == 1:
        string.append('There are no similar arguments.')

    return '\n\n'.join(string)



import argparse
import sys
import os
import yaml

from copy import deepcopy

from collections import UserDict
from pathlib import Path
from warnings import warn

from . import Namespacify, nested_dict_update
from .core import unflatten, get_similar_args_str_fmt
from .core._parse import ListType, ListAction, parse_arg_type, get_type
from .logging import get_logger
from .utils import api


DEFAULT_CONFIG_PATH = os.path.join(os.getcwd(), 'default_config.yaml')


class Config(Namespacify):
    """
    Serializable config object.

    Parameters
    ----------
    config : str, Path, dict, None or list of str, dict, or Path, default None
        Dictionaries of configuration values or str/Path pointing to YAML file(s) containing configuration values.
        * If str or Path, the value is treated as a YAML file which is then loaded to a dictionary, and then:
        * If dictionary, `Config` object will be updated to contain values in the dictionary. These values can be either
          nested or '.'-delimited, the latter of which will be nested before updating `Config` object.
        * If list of the above, `Config` object will be updated in corresponding order. In the case of duplicate keys,
          the value from the **last** entry will be utilized.
    default : str, dict or Path, default `os.path.join(os.getcwd(), 'default_config.yaml')`.
        Dict or path-like object from which to load default config.
        The corresponding nested dictionary defines the arguments in argparse. For example:
         >>> default = {'truck': {'wheels': 4, 'brand': 'toyota'}}
         will result in argparse arguments `truck.wheels` of type `int` and `truck.brand` of type str, with defaults
         of 4 and 'toyota', respectively.
    yaml_type_handling: {'ignore', 'warn', 'error'}, default 'warn'
        Handling method for values loaded from config files (either via `--config <filename>` or passed via `config`
        parameter) that cannot be cast to the type of the value in the default config.
        * 'ignore': silently allow mistyped values.
        * 'warn': allow mistyped values and raise a warning if encountered.
        * 'error': raise `TypeError`s on mistyped values.
    Attributes
    ----------

    """
    def __init__(self, config=None, default=DEFAULT_CONFIG_PATH, yaml_type_handling='warn'):
        assert yaml_type_handling in ('ignore', 'warn', 'error'), \
            "yaml_type_handling must be one of 'ignore', 'warn', error'"

        self.yaml_type_handling = yaml_type_handling

        self.default_config = DefaultConfig(self._parse_default(config, default))
        self.logger = get_logger()
        self.verbosity = 0

        super().__init__(self._parse_config())

        self.update_with_configs(config)
        self.verbose(self.verbosity)

    def _parse_default(self, config, default):
        if api.is_dict_like(default):
            return default

        candidates = [Path(default), (Path(sys.argv[0]).parent / default)]

        if config is not None and isinstance(config, (str, Path)):
            candidates.extend([Path(config), (Path(sys.argv[0]).parent / config)])

        for candidate in candidates:
            if candidate.exists():
                return candidate

        candidates_str = '\n\t'.join(str(x.absolute()) for x in candidates)
        err_msg = f'Attempted to resolve default config in the following order:\n\t{candidates_str}.\n' \
                  f'Unable to find a file amongst these candidates.'
        raise ValueError(err_msg)

    def _parse_config(self):
        # First we parse any --config arguments and load those
        # Then we can override them with any other passed values.
        base_config = deepcopy(self.default_config)

        config_file_args, other_args = self._split_config_file_args()
        config_files = self._create_config_file_parser().parse_args(args=config_file_args)
        self.update_with_configs(config_files.config, base_config)

        parsed_args = self._create_parser(default=base_config).parse_known_args(args=other_args)

        if len(parsed_args[1]):
            valid_option_keys = sorted(parsed_args[0].__dict__.keys())
            warn_msg = get_similar_args_str_fmt(parsed_args[1], valid_option_keys)
            warn(warn_msg)

        args_dict = self._extract_verbosity(parsed_args[0].__dict__)
        restructured = unflatten(args_dict)

        self._check_restructured(restructured, self.default_config)
        return restructured

    def _split_config_file_args(self):
        config_arg_loc = [x.startswith('--config') for x in sys.argv]

        if sum(config_arg_loc) > 1:
            msg = "Multiple --config arguments encountered. Pass multiple configs with a single --config prefix, e.g. " \
                  "'python script.py --config config_1.yaml config_2.yaml'"
            raise TypeError(msg)

        try:
            config_key_idx = sys.argv.index('--config')
        except ValueError:
            try:
                # single string starting with '--config'
                assert sum(config_arg_loc) == 1
            except AssertionError:
                return [], sys.argv[1:]
            else:
                config_arg_loc = config_arg_loc.index(True)
                config_arg = sys.argv[config_arg_loc]

                other_args = sys.argv[1:].copy()
                other_args.pop(config_arg_loc-1)

                return [config_arg], other_args

            return [], sys.argv[1:]

        config_args = []
        other_args = []
        try:
            next_optional_arg = [j for j in range(config_key_idx+1, len(sys.argv)) if '--' in sys.argv[j]][0]
        except IndexError:
            # no optional args following config arguments
            next_optional_arg = float('inf')

        for j, arg in enumerate(sys.argv[1:], start=1):
            if config_key_idx <= j < next_optional_arg:
                config_args.append(arg)
            else:
                other_args.append(arg)

        assert set(config_args).union(other_args) == set(sys.argv[1:])
        assert not set(config_args).intersection(other_args)

        return config_args, other_args

    def _create_parser(self, default=None):
        parser = argparse.ArgumentParser()
        for arg_name, arg_info in self._get_arguments(d=default).items():
            parser.add_argument(f'--{arg_name}', **arg_info)

        if parser.get_default('verbose') is None:
            parser.add_argument('--verbose', default=0, type=int)

        return parser

    def _create_config_file_parser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--config', default=[], nargs='+', type=ListType(str), action=ListAction)
        return parser

    def update_with_configs(self, configs, updatee=None):
        """
        Update `self` with one or multiple dict-like objects.

        Parameters
        ----------
        configs : str, Path, dict or None
            One of:
            * Path-like pointing to a yaml-file to load a dict from
            * dict
            * list of the above
            * None, in which case no update occurs
        updatee : dict-like or None, default None
            Object to update. If None, updates self.

        Returns
        -------
        updatee: dict-like
            Updated object.

        """
        if configs is None:
            return self
        elif not api.is_list_like(configs) or api.is_dict_like(configs):
            configs = [configs]

        for config in configs:
            updatee = self._update_with_config(config, updatee=updatee)

        return updatee

    def _update_with_config(self, config, updatee=None):
        if isinstance(config, (str, Path)):
            config = _config_from_yaml(config)

        config = self._restructure_as_necessary(config)

        if updatee:
            return nested_dict_update(updatee, config)
        else:
            return nested_dict_update(self, config)

    def _restructure_as_necessary(self, config):
        if any('.' in k for k in config.keys()):
            if any(isinstance(v, dict) for v in config.values()):
                raise ValueError('Cannot combine nested dict config arguments with "." deliminated arguments.')

            config = unflatten(config)

        return config

    def _extract_verbosity(self, config):
        self.verbosity = config['verbose']

        try:
            _ = self.default_config.verbose
        except AttributeError:
            config.pop('verbose')

        return config

    def _check_restructured(self, restructured, default_config, *stack):
        for key, value in default_config.items():
            if key not in restructured:
                raise RuntimeError(f'Missing key {"->".join([*stack, key])} in restructured config.')
            elif isinstance(value, dict):
                self._check_restructured(restructured[key], value, *stack, key)

    def _get_arguments(self, key='', d=None):
        if d is None:
            d = self.default_config

        args = {}

        for k, v in d.items():
            new_key = f'{key}.{k}' if key else k
            if isinstance(v, (dict, UserDict)) and len(v):
                args.update(self._get_arguments(key=new_key, d=v))
            else:
                if '-' in new_key:
                    raise NameError(f"Invalid character '-' in key '{new_key}'.")

                base_default = self.default_config.get(tuple(new_key.split('.')), v)
                args[new_key] = self._collect_argument(v, base_default, new_key)

        return args

    def _collect_argument(self, default_val, base_default, arg_name):
        arg = {}

        _type, additional_args = parse_arg_type(arg_name, base_default)

        arg.update(additional_args)

        try:
            default = _type(default_val)
        except Exception as e:
            msg = f"Value '{default_val}' read from yaml file cannot be case to type '{get_type(_type)}' "\
                  f"of base config value."
            if self.yaml_type_handling == 'error':
                raise TypeError(msg) from e
            elif self.yaml_type_handling == 'warn':
                self.logger.warning(msg)

            default = default_val

        arg.update({'default': default, 'type': _type})

        return arg


    def serialize_to_dir(self, log_dir, fname='config.yaml', use_existing_dir=False, with_default=False):
        """
        Save the config as a yaml file in a directory.

        Parameters
        ----------
        log_dir : str or Path
            directory to serialize into.

        fname : str, default 'config.yaml`
            Name of the file.

        use_existing_dir : bool, default False
            Whether to serialize to a directory that already exists. If False and `log_dir` exists,
            config will be serialized to `{log_dir}_{k}`, where k is the smallest positive integer such that
            `{log_dir}_{k}` does not currently exist.

        with_default : bool, default False
            Whether to serialize default config as well. If true, default config is serialized as `config_default.yaml`
            in the same `log_dir` as the config. The symmetric difference is also serialized as
            `config_difference.yaml`.

        Returns
        -------
        log_dir : str
            Path of the log dir the config was serialized to.

        """
        log_dir = super().serialize_to_dir(log_dir, fname=fname, use_existing_dir=use_existing_dir)

        if with_default:
            path = Path(fname)

            def fname_func(kind): return (path.parent / f'{path.stem}_{kind}').with_suffix(path.suffix)

            self.default_config.serialize_to_dir(log_dir,
                                                 fname=fname_func('default'),
                                                 use_existing_dir=True)

            (self ^ self.default_config).serialize_to_dir(log_dir,
                                                          fname=fname_func('difference'),
                                                          use_existing_dir=True)
        return log_dir

    def verbose(self, level):
        if level >= 2:
            self.logger.info('Trainer config:')
            self.pprint(indent=1, log_func=self.logger.info)
        if level >= 1:
            xor = self ^ self.default_config
            print(f'\n{"-"*10}\n')
            if xor:
                self.logger.info('Custom trainer config (difference from default):')
                xor.pprint(indent=1, log_func=self.logger.info)
            else:
                self.logger.info('No difference from default.')


class DefaultConfig(Namespacify):
    def __init__(self, default):
        if not api.is_dict_like(default):
            default = _config_from_yaml(default)

        super().__init__(default)


def _config_from_yaml(file_path):
    contents = Path(file_path).expanduser().open('r')
    loaded_contents = yaml.safe_load(contents)

    if not isinstance(loaded_contents, dict):
        raise ValueError(f'Contents of file "{file_path}" deserialize into object of type '
                         f'{type(loaded_contents).__name__}, should be dict.')

    return loaded_contents



import re
import sys
from io import StringIO, TextIOBase, TextIOWrapper
from contextlib import ContextDecorator
from enum import IntEnum
from typing import Union


class TapeRecorder(ContextDecorator):
    _dest: 'Union[TextIOWrapper, None]'
    _string: 'StringIO' = None
    _stdout: '_IOList' = None
    _stderr: '_IOList' = None

    def __init__(self, dest=None):
        super().__init__()
        self._dest = None
        self._status = TapeStatus.UNOPENED

        if dest is not None:
            self._init(dest)

    def _init(self, dest=None):
        self._dest = None
        self._string = StringIO()

        if dest is not None:
            self.add_file(dest)

        self._stdout = _IOList(sys.stdout, self.destination)
        self._stderr = _IOList(sys.stderr, self.destination)
        self._status = TapeStatus.INITIALIZED

    def reset(self, dest=None):
        if dest is not None:
            self.add_file(dest, copy_history=False)
        else:
            new_string = StringIO()
            self._replace_destination(new_string)

            self._dest = None
            self._string = new_string

        self._status = TapeStatus.INITIALIZED

    def add_file(self, file_like, copy_history=True):
        if not isinstance(file_like, TextIOBase):
            file_like = open(file_like, 'w')

        if self._status >= TapeStatus.INITIALIZED:
            if copy_history:
                file_like.write(self.read_tape())
                file_like.flush()

            if not self._string.closed:
                self._string.close()

        self.set_dest(file_like)

    def set_dest(self, dest):
        self._replace_destination(dest)
        self._dest = dest

    def _replace_destination(self, new_destination):
        if self._stdout is not None:
            self._stdout.replace(self.destination, new_destination)

        if self._stderr is not None:
            self._stderr.replace(self.destination, new_destination)

    def read_tape(self):
        if self._dest is None:
            try:
                return self._string.getvalue()
            except ValueError:
                raise ValueError('Attempting to read from never-entered tape.')

        with open(self._dest.name, 'r') as f:
            return f.read()

    @property
    def destination(self):
        return self._dest or self._string

    @destination.setter
    def destination(self, value):
        self.set_dest(value)

    @property
    def status(self):
        return self._status

    def record(self, dest=None, copy_history=True):
        self.__enter__(dest, copy_history)

    def end_record(self):
        self.__exit__(None, None, None)

    def __del__(self):
        self.__exit__(None, None, None)

    def __enter__(self, dest=None, copy_history=True):
        if self._status == TapeStatus.UNOPENED:
            self._init(dest)
        elif self._status == TapeStatus.CLOSED:
            self.reset(dest)
        elif dest is not None:
            self.add_file(dest, copy_history=copy_history)

        for cm in (self._stdout, self._stderr):
            cm.__enter__()

        self._status = TapeStatus.ACTIVE

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for cm in (self._stdout, self._stderr):
            cm.__exit__(exc_type, exc_val, exc_tb)

        self._status = TapeStatus.CLOSED


class _IOList(TextIOBase):
    def __init__(self, original_stream, *io_streams):
        self._original_stream = original_stream
        self._io_streams = set(io_streams)

        self._original_write = self._original_stream.write
        self._original_flush = self._original_stream.flush

    def write(self, line):
        o = [self._original_write(line)] + [s.write(escape_ansi(line)) for s in self._io_streams]
        return max(o)

    def flush(self):
        self._original_flush()

        for s in self._io_streams:
            s.flush()

    def close(self):
        for s in self._io_streams:
            s.close()

    def add(self, stream):
        self._io_streams.add(stream)

    def remove(self, stream):
        self._io_streams.remove(stream)

    def replace(self, old, new):
        self._io_streams.remove(old)
        self._io_streams.add(new)

    def __contains__(self, item):
        return self._io_streams.__contains__(item)

    def __enter__(self):
        self._original_stream.write = self.write
        self._original_stream.flush = self.flush
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._original_stream.write = self._original_write
        self._original_stream.flush = self._original_flush
        self.close()


class TapeStatus(IntEnum):
    UNOPENED = -2
    INITIALIZED = -1
    ACTIVE = 0
    CLOSED = 1


def escape_ansi(line):
    ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', line)



from expfig import Namespacify
from expfig.utils.get_pandas import pandas as pd
from expfig.utils.api import is_dict_like


def compare(ns, base=None):
    """
    Parameters
    ----------
    ns: dict or list-like of Namespacify or path-like.
        Either Namespacifys to compare or paths to yaml files to load namespacifys from.
    base: Namespacify, path-like or None, default None.
        Base namespacify to compare to. If None, only pairwise comparisons are performed. Otherwise, both pairwise and
        comparisons to the base are performed.

    Returns
    -------

    """
    if base is not None:
        base = _load_if_necessary(base)

    if not is_dict_like(ns):
        ns = dict(enumerate(ns))

    ns = {str(k): _load_if_necessary(v) for k, v in ns.items()}

    pairwise_diffs = {}
    for name, namespace in ns.items():
        diffs = []
        if base is not None:
            diffs.append(namespace.difference(base).to_series())

        diffs.extend(namespace.difference(other_namespace).to_series() for other_namespace in ns.values())
        pairwise_diffs[name] = pd.concat(diffs).drop_duplicates()

    return pd.concat(pairwise_diffs, axis=1).sort_index()


def _load_if_necessary(value):
    if isinstance(value, Namespacify):
        return value

    return Namespacify.from_yaml(value)



from ._flatten import depth, unflatten, flatten, nested_dict_update
from ._str_types import str2bool, str2none, TypeToNone
from .similar_args import get_similar_args_str_fmt
from ._parse import get_type



import logging
import os
import tempfile
from expfig.logging import get_logger


def make_sequential_log_dir(log_dir, subdirs=(), use_existing_dir=False, logger_file=None, logger_level=None):
    """Taken from rlworkgroup/garage.

    Creates log_dir, appending a number if necessary.

    Attempts to create the directory `log_dir`. If it already exists, appends
    "_1". If that already exists, appends "_2" instead, etc.

    If `logger_file` is not None, a logger will be created that logs to the file `logger_file` in the created `log_dir`.
    This logger can be accessed by calling `expfig.logging.get_logger().

    Args:
        log_dir (str or None): The log directory to attempt to create. If None, logs to a temporary directory
        subdirs (list of str): subdirectories to create in the log_dir directory.
        use_existing_dir (bool): whether to simply return the dir if it exists (will log to existing dir).
        logger_file (str or None): If not None, creates a logger that logs to this file within the log_dir.
        logger_level: (int, str, or None): Level for the logger, e.g. `INFO`. See python Logging module for more info.
            If None and logger_file is not None, logger_level will be set to logging.DEBUG.
            If None and logger_file is None, no logger will be created.

    Returns:
        str: The log directory created.

    """

    if log_dir is None:
        log_dir = tempfile.mkdtemp()

    i = 0
    while True:
        try:
            if i == 0:
                os.makedirs(log_dir)
            else:
                possible_log_dir = '{}_{}'.format(log_dir, i)
                os.makedirs(possible_log_dir)
                log_dir = possible_log_dir

            break

        except FileExistsError:
            if use_existing_dir:
                break

            i += 1

    for subdir in subdirs:
        os.makedirs(os.path.join(log_dir, subdir), exist_ok=True)


    if logger_file is not None:
        logger_file = os.path.join(log_dir, logger_file)
        if logger_level is None:
            logger_level = logging.DEBUG

    if logger_level is not None:
        logger = get_logger(level=logger_level, log_file=logger_file)

    return log_dir



class BadPandas:
    def __getattr__(self, item):
        if item.startswith('__'):
            # Alleviates pytest issue with self.__unwrapped__, see https://github.com/pytest-dev/pytest/issues/5080
            raise AttributeError(item)

        raise ModuleNotFoundError("Optional dependency 'pandas' not installed. "
                                  "Install with 'pip install pandas' to utilize this functionality.")


def _get_pandas():
    try:
        import pandas as pd
    except ImportError:
        return BadPandas()
    else:
        return pd


pandas = _get_pandas()



import pytest
import yaml

from expfig.core._parse_yaml_obj import YamlType


class AYamlObject(yaml.YAMLObject):
    yaml_dumper = yaml.SafeDumper
    yaml_loader = yaml.SafeLoader
    yaml_tag = u'!AYamlObject'

    def __init__(self, value):
        self.value = value


class TestYamlType:
    def test_yaml_default_is_yaml(self):
        yaml_type = YamlType(yaml_default=True)

        loaded = yaml_type('!AYamlObject {value: 10}')

        assert isinstance(loaded, AYamlObject)
        assert loaded.value == 10

    def test_yaml_default_is_not_yaml(self):
        yaml_type = YamlType(yaml_default=True)

        with pytest.raises(yaml.YAMLError):
            _ = yaml_type('abcd')

    def test_yaml_default_is_bad_yaml(self):
        yaml_type = YamlType(yaml_default=True)

        with pytest.raises(yaml.YAMLError):
            _ = yaml_type('!MisspelledYamlObject {value: 10}')



import pytest

from expfig.core._parse import ListType
from expfig.core import str2none


class TestListTypeFromList:
    def test_all_str(self):
        list_like = 'a', 'b', 'c'
        list_type = ListType.from_list(list_like)

        assert list_type.type == str

    def test_all_int(self):
        list_like = 0, 4, 10
        list_type = ListType.from_list(list_like)

        assert list_type.type == int

    def test_all_float(self):
        list_like = 0.0, 4.5, 10.0
        list_type = ListType.from_list(list_like)

        assert list_type.type == float

    def test_nonunique(self):
        list_like = 'a', 0

        with pytest.warns(UserWarning, match='Collecting list-like argument'):
            list_type = ListType.from_list(list_like)

        assert list_type.type == str



import argparse

from ast import literal_eval
from warnings import warn

from expfig.core import str2bool, str2none, TypeToNone
from expfig.core._parse_yaml_obj import YamlType
from expfig.utils import api


def parse_arg_type(arg_name, base_default):
    additional_args = {}

    if api.is_list_like(base_default):
        additional_args.update(nargs='+', action=ListAction)
        _type = ListType.from_list(base_default, arg_name)

    elif getattr(base_default, 'yaml_tag', None) is not None:  # is a Yaml object
        _type = YamlType(yaml_default=True)
    elif base_default is None or base_default == '': # allow empty strings to be yaml-objects but no failure if not
        _type = YamlType(yaml_default=False)
    else:
        _type = type(base_default)

    if _type == bool:
        _type = str2bool
    elif _type == str:
        _type = str2none

    return _type, additional_args


class ListType:
    __name__ = 'ListType'
    # TODO (ahalev) make this a metaclass, inherit from type and deprecate this class attr

    def __init__(self, _type):
        self._type = TypeToNone(_type)

    def __call__(self, value):
        if isinstance(value, list):
            literal = [self._type(v) for v in value]
        elif self._type.type == str:
            literal = self.str2none_eval(value)
        else:
            try:
                return self._type(value)
            except ValueError:
                literal = literal_eval(value)
                return self(literal)

        return self.type_check(literal)

    def type_check(self, value):
        if isinstance(value, list) and all(isinstance(v, self.valid_types) for v in value) or \
                isinstance(value, self.valid_types):
            return value

        raise argparse.ArgumentTypeError(f'Invalid value(s) for type {self._type}: {value}')

    @property
    def valid_types(self):
        return self._type.valid_types

    @property
    def type(self):
        return self._type.type

    @staticmethod
    def str2none_eval(value):
        if value.startswith('[') and value.endswith(']'):
            _value = literal_eval(value)
            return [str2none(v) for v in _value]

        return str2none(value)

    @classmethod
    def from_list(cls, list_like, arg_name=None):
        unique_types = {type(x) for x in list_like if x is not None}

        if len(unique_types) == 1:
            _type = unique_types.pop()
        else:
            _type = str

            if len(unique_types):
                arg = f"'{arg_name}' " if arg_name else ""
                warn(f"Collecting list-like argument {arg}with non-unique types {unique_types} in default value "
                     "Collected values will be str.")

        return cls(_type)


class ListAction(argparse._StoreAction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if len(values) == 1 and isinstance(values[0], list) and self.type.type != list:
            values = values[0]

        return super().__call__(parser, namespace, values, option_string=None)


def get_type(type_obj):
    try:
        return type_obj.type.__name__
    except AttributeError:
        return type_obj.__name__



import functools

import numpy as np
import yaml

from contextlib import contextmanager
from collections import UserDict
from logging import getLogger

from . import nested_dict_update
from .core import depth, flatten
from .logging import make_sequential_log_dir

from expfig.utils.api import is_dict_like
from expfig.utils.get_pandas import pandas as pd


yaml.SafeDumper.add_multi_representer(UserDict, yaml.SafeDumper.represent_dict)
logger = getLogger(__name__)

class Namespacify(UserDict):
    def __init__(self, in_dict):
        super().__init__(in_dict)

    def update(self, *args, **kwargs):
        return nested_dict_update(self, *args, nest_namespacify=True, **kwargs)

    def pprint(self, indent=0, log_func=None, _recur=False):
        with repr_as_default_yaml_representer():
            str_out = yaml.safe_dump(self)

        if indent:
            pref = '\t'*indent
            str_out = str_out.replace('\n', '\n' + pref)
            str_out = pref + str_out

        if log_func is not None:
            log_func(str_out)
        else:
            print(str_out)

        return str_out

    def depth(self):
        return depth(self)

    def to_dict(self):
        return {k: v.to_dict() if isinstance(v, Namespacify) else (v.copy() if hasattr(v, 'copy') else v)
                for k, v in self.items()}

    def to_series(self):
        series = pd.json_normalize(self.to_dict()).squeeze()

        try:
            series.index = pd.MultiIndex.from_tuples([x.split('.') for x in series.index])
        except TypeError:
            if not series.empty:
                raise

        return series

    def flatten(self, delimiter='.', levels=None):
        return flatten(self, delimiter=delimiter, levels=levels)

    def intersection(self, other):
        intersection = {}

        for k, v in self.items():
            if k in other:
                if equal(other[k], v):
                    intersection[k] = v
                elif isinstance(v, Namespacify) and isinstance(other[k], Namespacify):
                    subint = v.intersection(other[k])
                    if subint:
                        intersection[k] = subint

        return Namespacify(intersection)

    def symmetric_difference(self, other):
        """
        Get all values that differ in ``self`` or ``other``.

        Returns the value in ``self`` if it exists and differs from ``other``. Otherwise returns the value in ``other``.

        If ``self['a'] = 1`` and ``other['a'] = 2``, ``self.symmetric_difference(other)['a'] = 1``.
        If ``self['a'] = 1`` and ``'a' not in other``, `self.symmetric_difference(other)['a'] = 1``.
        If ``'a' not in self`` and ``other['a'] = 2``, `self.symmetric_difference(other)['a'] = 2``.


        Parameters
        ----------
        other : dict-like
            Object to compare against

        Returns
        -------
        difference : :class:`.Namespacify`
            Difference between ``self`` and ``other``.

        """

        diff = {}

        if type(self) != type(other):
            return self.copy()

        keys = {*self.keys(), *other.keys()}
        for k in keys:
            if k not in self:
                diff[k] = other[k]
                continue

            elif k not in other:
                diff[k] = self[k]
                continue

            elif not equal(self[k], other[k]):
                if isinstance(self[k], Namespacify):
                    diff[k] = self[k].symmetric_difference(other[k])
                else:
                    diff[k] = self[k]

        return Namespacify(diff)

    def difference(self, other):
        """
        Get all values that are in ``self`` that are NOT (or are different) in ``other``.

        If ``self['a'] = 1`` and ``other['a'] = 2``, ``self.difference(other)['a'] = 1``.
        If ``a not in self`` and ``other['a'] = 2``, ``self.difference(other)['a']`` returns a ``KeyError``.

        Parameters
        ----------
        other : dict-like
            Object to compare against

        Returns
        -------
        difference : :class:`.Namespacify`
            Difference between ``self`` and ``other``.

        """
        diff = {}
        for k, v in self.items():
            if not is_dict_like(other):
                diff[k] = v
            elif k not in other:
                diff[k] = v
            elif not equal(v, other[k]):
                if isinstance(v, Namespacify):
                    diff[k] = v.difference(other[k])
                else:
                    diff[k] = v

        return Namespacify(diff)

    def serialize(self, stream=None):
        return yaml.safe_dump(self, stream=stream)

    def serialize_to_dir(self, log_dir, fname='namespacify.yaml', use_existing_dir=False):
        log_dir = make_sequential_log_dir(log_dir, use_existing_dir=use_existing_dir)
        log_file = f'{log_dir}/{fname}'

        with open(log_file, 'w') as f:
            self.serialize(f)

        logger.info(f'Logged {type(self).__name__} to {log_file}')

        return log_dir

    @classmethod
    def deserialize(cls, stream):
        return cls(yaml.safe_load(stream))

    @classmethod
    def from_yaml(cls, filepath):
        with open(filepath, 'r') as stream:
            return cls.deserialize(stream)

    def __dir__(self):
        rv = set(super().__dir__())

        try:
            rv = rv | set(self.keys())
        except RuntimeError:
            pass

        return sorted(rv)

    def __getitem__(self, item):
        if item[0] == slice(None) or isinstance(item[0], list):
            keys = self.keys() if item[0] == slice(None) else item[0]
            return {k: self[k][item[1:]] for k in keys}

        elif isinstance(item, tuple):
            out = self[item[0]]
            if len(item) == 1:
                return out
            try:
                return out[item[1:]]
            except TypeError:  # out is not subscriptable, raise KeyError
                raise KeyError(f"'{item[1:]}', item of type '{type(out).__name__}' is not subscriptable")

        return super().__getitem__(item)

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            nested_update = functools.reduce(lambda val, k: {k: val}, reversed(key), value)
            nested_dict_update(self, nested_update, nest_namespacify=True)

        elif isinstance(value, dict):
            nested_dict_update(self, {key: value}, nest_namespacify=True)

        else:
            super().__setitem__(key, value)

    def __getattr__(self, item):
        if item == 'data':
            raise RuntimeError('Attempting to access self.data before initialization.')
        try:
            return self[item]
        except (KeyError, RuntimeError):
            raise AttributeError(item)

    def __setattr__(self, key, value):
        try:
            contains_key = key in self
        except RuntimeError:
            pass
        else:
            if contains_key:
                self[key] = value
                return

        super().__setattr__(key, value)

    def __xor__(self, other):
        return self.symmetric_difference(other)

    def __and__(self, other):
        return self.intersection(other)

    def __sub__(self, other):
        return self.difference(other)

    def __deepcopy__(self, memo=None):
        return Namespacify(self.to_dict())


def equal(a, b):
    try:
        return bool(a == b)
    except ValueError:
        return np.array_equal(a, b)


@contextmanager
def repr_as_default_yaml_representer():
    def default_representer(dumper, data):
        return dumper.represent_scalar('tag:yaml.org,2002:str', repr(data))

    yaml.representer.SafeRepresenter.add_representer(None, default_representer)

    try:
        yield
    finally:
        yaml.representer.SafeRepresenter.yaml_representers.pop(None)



from pathlib import Path
from setuptools import setup, find_packages

DESCRIPTION = 'A yaml-based configuration for reproducible python experiments.'
VERSION = '1.0.2'
MAINTAINER = 'Avishai Halev'
MAINTAINER_EMAIL = 'avishaihalev@gmail.com'
LICENSE = 'MIT License'
PROJECT_URLS = {
    'Source Code': f'https://github.com/ahalev/experiment-config/tree/v{VERSION}'
}

EXTRAS = {
    'dev': ['numpy', 'pytest'],
    'pandas': ['pandas'],
}

EXTRAS['all'] = sum(EXTRAS.values(), [])


setup(
    name='experiment-config',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    python_requires='>=3.6',
    version=VERSION,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    project_urls=PROJECT_URLS,
    description=DESCRIPTION,
    license=LICENSE,
    long_description=(Path(__file__).parent / 'README.md').read_text(),
    long_description_content_type='text/markdown',
    include_package_data=True,
    install_requires=[
        'pyyaml'
    ],
    extras_require=EXTRAS

)



import contextlib
import enum
import functools
import logging
import os
import sys


def get_logger(level=logging.DEBUG, log_file=None):
    logger = logging.getLogger(__name__)
    logger.setLevel(level)

    formatter = ColorFormatter('%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s')

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level)
    handler.setFormatter(formatter)

    if not check_existing_handler(handler, logger.handlers):
        logger.addHandler(handler)

    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        if not check_existing_handler(file_handler, logger.handlers):
            logger.addHandler(file_handler)
            logger.info(f'Logging to file: {os.path.abspath(log_file)}')

    return logger

class ColorFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.formats = Formats.level_formats(self._style._fmt)

    def format(self, record: logging.LogRecord) -> str:
        with self.push_format(record.levelno):
            return super().format(record)

    @contextlib.contextmanager
    def push_format(self, level):
        old_fmt = self._style._fmt
        self._style._fmt = self.formats[level]

        try:
            yield
        finally:
            self._style._fmt = old_fmt


class Formats(enum.Enum):
    GRAY = "\x1b[38;20m"
    YELLOW = "\x1b[33;20m"
    RED = "\x1b[31;20m"
    BOLD_RED = "\x1b[31;1m"
    RESET = "\x1b[0m"

    LEVELS = {
        logging.DEBUG: GRAY,
        logging.INFO: GRAY,
        logging.WARNING: YELLOW,
        logging.ERROR: RED,
        logging.CRITICAL: BOLD_RED
    }

    @classmethod
    def level_formats(cls, format_str):
        return {level: cls.format(format_str, color) for level, color in cls.LEVELS.value.items()}

    @classmethod
    def format(cls, format_str, color):
        return ''.join([color, format_str, cls.RESET.value])


def check_existing_handler(new_handler, existing_handlers):
    for handler in existing_handlers:
        if handler.stream == new_handler.stream:
            return True

    return False
